#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys

from typing import Optional

from environs import Env
from loguru import logger


def setup_logger(level: str = "INFO", env: Optional[Env] = None) -> int:
    force_debug = env.bool("PRAMENPY_DEBUG", False) if env else False
    level = "DEBUG" if force_debug else level
    logger.remove()
    if level == "INFO":
        logger.disable("py4j")
    logger_id = logger.add(
        sink=sys.stderr,
        level=level,
        format="{time} | {level} | {name} | {extra} | {message}",
    )
    logger.debug(f"Logger setup complete. Logger level is {level}.")
    return logger_id



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import enum
import inspect

from typing import Any, Callable, Dict, Generic, Tuple, TypeVar
from unittest.mock import MagicMock

import pytest

from pytest_mock import MockerFixture
from typing_extensions import ParamSpec, Protocol


class HasNameDunder(Protocol):
    __name__: str


_TargetClsType = TypeVar("_TargetClsType", bound=HasNameDunder)
_TargetMethodParams = ParamSpec("_TargetMethodParams")
_TargetMethodReturn = TypeVar("_TargetMethodReturn")
_TargetMethodKey = Tuple[str, str]

_TargetMethodArgs = Tuple[Any, ...]
_TargetMethodKwargs = Dict[str, Any]

_CallKeyParamDef = Tuple[str, Any]
_CallKey = Tuple[_CallKeyParamDef, ...]


class Markers(enum.Enum):
    """Markers for defining When.called_with arguments.

    Markers.any - means the argument could be anything
    """

    any: str = "any"


def create_call_key(
    original_callable_sig: inspect.Signature,
    *args: _TargetMethodArgs,
    **kwargs: _TargetMethodKwargs,
) -> _CallKey:
    """Create a call identification key for the given function.

    This key represents a certain call to the function. The order
    of kwargs is not important and will be allocated based on order
    of the kwargs in the function signature.

    Supports normal functions as well as class methods.
    """
    original_params = original_callable_sig.parameters
    args_key: _CallKey = tuple(
        zip(tuple(original_params)[: len(args)], args),
    )

    def get_from_kwargs_or_default(k: str) -> Any:
        may_be_result = kwargs.get(k, original_params[k].default)
        if may_be_result is inspect.Parameter.empty:
            raise ValueError(
                f"Not specified parameter {k}. "
                f"Incompatible call specification args={args}, kwargs={kwargs}"
                f" to the function with signature {original_callable_sig}"
            )
        else:
            return may_be_result

    kwargs_key: _CallKey = tuple(
        (k, get_from_kwargs_or_default(k))
        for k in tuple(original_params)[len(args) :]
    )
    return args_key + kwargs_key


def get_mocked_call_result(
    original_callable_sig: inspect.Signature,
    mocked_calls: Dict[
        _CallKey,
        _TargetMethodReturn,
    ],
    *args: _TargetMethodArgs,
    **kwargs: _TargetMethodKwargs,
) -> _TargetMethodReturn:
    call_key = create_call_key(
        original_callable_sig,
        *args,
        **kwargs,
    )

    def params_are_compatible(
        param_in_mocked_calls: _CallKeyParamDef,
        param_in_call: _CallKeyParamDef,
    ) -> bool:
        assert param_in_mocked_calls[0] == param_in_call[0]
        return (
            param_in_mocked_calls[1] is Markers.any
            or param_in_mocked_calls[1] == param_in_call[1]
        )

    def call_matched_call_key(mocked_call_key: _CallKey) -> bool:
        return all(
            params_are_compatible(param_1, param_2)
            for param_1, param_2 in zip(mocked_call_key, call_key)
        )

    for call in filter(call_matched_call_key, mocked_calls):
        return mocked_calls[call]
    else:
        raise KeyError(
            f"Call {call_key} is not in mocked_calls {mocked_calls}"
        )


def side_effect_factory(
    origin_callable: Callable[_TargetMethodParams, _TargetMethodReturn],
    mocked_calls: Dict[_CallKey, _TargetMethodReturn],
) -> Callable[_TargetMethodParams, _TargetMethodReturn]:
    def side_effect(
        *args: _TargetMethodParams.args,
        **kwargs: _TargetMethodParams.kwargs,
    ) -> _TargetMethodReturn:
        try:
            return get_mocked_call_result(
                inspect.signature(origin_callable),
                mocked_calls,
                # TODO investigate why mypy failing
                *args,
                **kwargs,
            )
        except KeyError:
            return origin_callable(*args, **kwargs)

    return side_effect


class MockedCalls(
    Generic[
        _TargetClsType,
        _TargetMethodParams,
        _TargetMethodReturn,
    ]
):
    mocked_calls_registry: Dict[
        _TargetMethodKey,
        Dict[_CallKey, _TargetMethodReturn],
    ] = {}

    def __init__(self, mocker: MockerFixture) -> None:
        self.mocker = mocker

    def add_call(
        self,
        cls: _TargetClsType,
        method: str,
        args: _TargetMethodArgs,
        kwargs: _TargetMethodKwargs,
        should_return: _TargetMethodReturn,
    ) -> MagicMock:
        self.mocked_calls_registry.setdefault(
            (cls.__name__, method),
            {},
        )
        self.mocked_calls_registry[(cls.__name__, method)][
            create_call_key(
                inspect.signature(getattr(cls, method)),
                *args,
                **kwargs,
            )
        ] = should_return

        return self.mocker.patch.object(
            cls,
            method,
            autospec=True,
            # it is important to send the origin target to the
            # side_effect_factory in order the result side_effect stores
            # the original target.
            side_effect=side_effect_factory(
                getattr(cls, method),
                self.mocked_calls_registry[(cls.__name__, method)],
            ),
        )


class When(
    Generic[
        _TargetClsType,
        _TargetMethodParams,
        _TargetMethodReturn,
    ]
):
    """Patching utility focused on readability.

    Example:

    >>> class Klass1:
    >>>     def some_method(
    >>>         self,
    >>>         arg1: str,
    >>>         arg2: int,
    >>>         *,
    >>>         kwarg1: str,
    >>>         kwarg2: str,
    >>>     ) -> str:
    >>>         return "Not mocked"


    >>> def test_should_properly_patch_calls(when):
    >>>     p = when(Klass1, "some_method").called_with(
    >>>         "a",
    >>>         Markers.any,
    >>>         kwarg1="b",
    >>>         kwarg2=Markers.any,
    >>>     ).then_return("Mocked")
    >>>
    >>>     assert (
    >>>         Klass1().some_method(
    >>>             "a",
    >>>             1,
    >>>             kwarg1="b",
    >>>             kwarg2="c",
    >>>         )
    >>>         == "Mocked"
    >>>     )
    >>>     assert (
    >>>         Klass1().some_method(
    >>>             "not mocked param",
    >>>             1,
    >>>             kwarg1="b",
    >>>             kwarg2="c",
    >>>         )
    >>>         == "Not mocked"
    >>>     )
    >>>     p.assert_called()

    It is possible to use 'when' with class methods and standalone functions
    (in this case cls parameter will become the python module).

    You can patch multiple times the same object with different "called_with"
    parameters in a single test.

    You can also patch multiple targets (cls, method)
    """

    cls: _TargetClsType
    method: str

    args: _TargetMethodArgs
    kwargs: _TargetMethodKwargs

    def __init__(self, mocker: MockerFixture):
        self.mocker = mocker
        self.mocked_calls = MockedCalls[
            _TargetClsType,
            _TargetMethodParams,
            _TargetMethodReturn,
        ](self.mocker)

    def __call__(
        self,
        cls: _TargetClsType,
        method: str,
    ) -> "When":  # type: ignore
        if inspect.ismethod(getattr(cls, method)):
            raise NotImplementedError(
                "See https://github.com/python/cpython/issues/67267 "
                "for more information"
            )

        def match_current_obj(patch) -> bool:  # type: ignore
            return patch.target is cls and patch.attribute == method

        # if current object was already patched, we have to re-patch it again
        for mocked_obj in filter(
            match_current_obj,
            self.mocker._patches,
        ):
            mocked_obj.stop()
            self.mocker._patches.remove(mocked_obj)

        self.cls = cls
        self.method = method
        return self

    def called_with(  # type: ignore
        self,
        *args,
        **kwargs,
    ) -> "When":  # type: ignore
        is_instance_method = (
            tuple(
                inspect.signature(
                    getattr(
                        self.cls,
                        self.method,
                    )
                ).parameters
            )[0]
            == "self"
        )
        # prepend Markers.any in case of a method (for self arg)
        self.args = (Markers.any,) + args if is_instance_method else args
        self.kwargs = kwargs
        return self

    def then_return(self, value: _TargetMethodReturn) -> MagicMock:
        return self.mocked_calls.add_call(
            self.cls,
            self.method,
            self.args,
            self.kwargs,
            value,
        )


@pytest.fixture
def when(mocker: MockerFixture) -> When:  # type: ignore
    """Patching utility focused on readability.

    Example:

    >>> class Klass1:
    >>>     def some_method(
    >>>         self,
    >>>         arg1: str,
    >>>         arg2: int,
    >>>         *,
    >>>         kwarg1: str,
    >>>         kwarg2: str,
    >>>     ) -> str:
    >>>         return "Not mocked"


    >>> def test_should_properly_patch_calls(when):
    >>>     when(Klass1, "some_method").called_with(
    >>>         "a",
    >>>         Markers.any,
    >>>         kwarg1="b",
    >>>         kwarg2=Markers.any,
    >>>     ).then_return("Mocked")

    >>>     assert (
    >>>         Klass1().some_method(
    >>>             "a",
    >>>             1,
    >>>             kwarg1="b",
    >>>             kwarg2="c",
    >>>         )
    >>>         == "Mocked"
    >>>     )
    >>>     assert (
    >>>         Klass1().some_method(
    >>>             "not mocked param",
    >>>             1,
    >>>             kwarg1="b",
    >>>             kwarg2="c",
    >>>         )
    >>>         == "Not mocked"
    >>>     )

    It is possible to use 'when' with class methods and standalone functions
    (in this case cls parameter will become the python module).

    You can patch multiple times the same object with different "called_with"
    parameters in a single test.

    You can also patch multiple targets (cls, method)
    """
    return When(mocker)



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import datetime
import enum
import pprint

from typing import Dict, List, Optional

import attrs
import cattr
import yaml

from click import Context, Option
from loguru import logger

from pramen_py.app import env
from pramen_py.utils import DEFAULT_DATE_FMT, convert_str_to_date


cattr.register_structure_hook(datetime.date, lambda v, t: v)  # type: ignore
cattr.register_unstructure_hook(datetime.date, lambda v: str(v))


@enum.unique
class TableFormat(enum.Enum):
    """The string name of each format should match Spark format"""

    parquet: str = "parquet"
    delta: str = "delta"


@attrs.define(auto_attribs=True, frozen=True, slots=True)
class InfoDateSettings:
    column: str = attrs.field()
    format: str = attrs.field(default="yyyy-MM-dd")
    start: Optional[datetime.date] = attrs.field(default=None)

    def __attrs_post_init__(self) -> None:
        # converter to datetime based on format
        if self.start and isinstance(self.start, str):
            logger.debug(
                f"Received {InfoDateSettings} with string as a date. "
                f"Converting {self.start} to the date with the "
                f"following format {self.format}"
            )
            object.__setattr__(
                self,
                "start",
                convert_str_to_date(self.start, self.format),
            )


@attrs.define(auto_attribs=True, frozen=True, slots=True)
class MetastoreTable:
    name: str = attrs.field()
    format: TableFormat = attrs.field()
    info_date_settings: InfoDateSettings = attrs.field()
    path: str = attrs.field(default="")
    table: str = attrs.field(default="")
    description: str = attrs.field(default="")
    records_per_partition: int = attrs.field(default=500000)
    reader_options: Dict[str, str] = attrs.field(factory=dict)
    writer_options: Dict[str, str] = attrs.field(factory=dict)

    def __attrs_post_init__(self) -> None:
        default_fs = env.str("PRAMENPY_DEFAULT_FS", "hdfs").strip('"')
        if default_fs != "local" and self.path.startswith("/"):
            logger.debug(
                f"Filesystem is {default_fs}, but path is {self.path}. "
                f"Appending {default_fs}:// prefix."
            )
            object.__setattr__(
                self,
                "path",
                f"{default_fs}://{self.path}",
            )


@attrs.define(auto_attribs=True, frozen=True, slots=True)
class RunTransformer:
    name: str = attrs.field()
    info_date: datetime.date = attrs.field()
    output_table: str = attrs.field()
    options: Dict[str, str] = attrs.field(factory=dict)
    spark_config: Dict[str, str] = attrs.field(factory=dict)

    def __attrs_post_init__(self) -> None:
        # converter to datetime
        if self.info_date and isinstance(self.info_date, str):
            logger.debug(
                f"Received {RunTransformer} with string as a date. "
                f"Converting {self.info_date} to the date with the "
                f"default format: {DEFAULT_DATE_FMT}"
            )
            object.__setattr__(
                self,
                "info_date",
                convert_str_to_date(
                    self.info_date,
                    DEFAULT_DATE_FMT,
                ),
            )


@attrs.define(auto_attribs=True, frozen=True, slots=True)
class TransformationConfig:
    run_transformers: List[RunTransformer] = attrs.field()
    metastore_tables: List[MetastoreTable] = attrs.field()


def serialization_callback(
    _: Context,
    __: Option,
    value: str,
) -> TransformationConfig:
    logger.info(
        f"Serializing and type checking the configuration received from "
        f"{value}"
    )
    with open(value) as config_f:
        config: Dict[str, str] = yaml.load(config_f, Loader=yaml.BaseLoader)
    try:
        serialized_config = cattr.structure(config, TransformationConfig)
        logger.info(pprint.pformat(cattr.unstructure(serialized_config)))
        return serialized_config
    except Exception as err:
        raise TypeError(f"Config validation failed:\n{err}") from err
